#include "datastructures.h"
#include "globals.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>

extern int overshoot;

void delete_database(database *D) {
	assert(D != NULL);
	if (D->T != NULL) {
		my_munmap(D->T, D->n + overshoot);
		D->T = NULL;
	}
	close_unlink(D->t_fd, D->t_filename);
	if (D->SA != NULL) {
		my_munmap(D->SA, D->n * sizeof(int));
		D->SA = NULL;
	}
	close_unlink(D->sa_fd, D->sa_filename);
	if (D->lcp != NULL) {
		my_munmap(D->lcp, D->n * sizeof(int));
		D->lcp = NULL;
	}
	close_unlink(D->lcp_fd, D->lcp_filename);
	if (D->a != NULL) {
		my_munmap(D->a, D->n * sizeof(int));
		D->a = NULL;
	}
	close_unlink(D->a_fd, D->a_filename);
	if (D->b != NULL) {
		my_munmap(D->b, D->n * sizeof(int));
		D->b = NULL;
	}
	close_unlink(D->b_fd, D->b_filename);
}

void close_database(database *D) {
	assert(D != NULL);
	if (D->T != NULL) {
		my_msync(D->T, D->n + overshoot);
		my_munmap(D->T, D->n + overshoot);
		D->T = NULL;
	}
	my_close(D->t_fd);
	if (D->SA != NULL) {
		my_msync(D->SA, D->n * sizeof(int));
		my_munmap(D->SA, D->n * sizeof(int));
		D->SA = NULL;
	}
	my_close(D->sa_fd);
	if (D->lcp != NULL) {
		my_msync(D->lcp, D->n * sizeof(int));
		my_munmap(D->lcp, D->n * sizeof(int));
		D->lcp = NULL;
	}
	my_close(D->lcp_fd);
	if (D->a != NULL) {
		my_msync(D->a, D->n * sizeof(int));
		my_munmap(D->a, D->n * sizeof(int));
		D->a = NULL;
	}
	my_close(D->a_fd);
	if (D->b != NULL) {
		my_msync(D->b, D->n * sizeof(int));
		my_munmap(D->b, D->n * sizeof(int));
		D->b = NULL;
	}
	my_close(D->b_fd);
}

void print_results(database *D) {
	int i, j;
	assert(D != NULL);
	for (i=0; i<D->n; ++i)
		for (j=D->a[i]; j<=D->b[i]; ++j) {
			fwrite(&D->T[D->SA[i]], j, 1, stdout);
			puts("");
		}
}

void create_childtab(database *D) {
	int last_index, *S, i, l;/*, maxstacksize, stacksize;*/
	assert(D != NULL && D->childtab != NULL && D->lcp != NULL);
	S = my_malloc(sizeof(int) * D->n);
	S[0] = 0;
	l = 1;
/*	maxstacksize = stacksize = l;*/
	last_index = -1;
	for (i=1; i<D->n; ++i) {
		while (D->lcp[i] < D->lcp[S[l-1]]) {
			last_index = S[--l];
			assert(l > 0);
			/* we are currently processing interval [S[l-1], i-1].
			   In the case D->lcp[i] == D->lcp[S[l-1]]
			   childtab[S[l-1]].nextl will be used,
			   and childtab[i-1].up will contain last_index */
			if (D->lcp[S[l-1]] != D->lcp[last_index]) {
/*				--stacksize;*/
				if (D->lcp[i] < D->lcp[S[l-1]]) {
					assert(D->lcp[S[l-1]] < D->lcp[last_index]);
					D->childtab[S[l-1]].down = last_index;
				}
			}
		}
		if (last_index != -1) {
			D->childtab[i-1].up = last_index;
			last_index = -1;
		}
		if (D->lcp[i] == D->lcp[S[l-1]])
			D->childtab[S[l-1]].nextl = i;
/*		else {
			++stacksize;
			if (stacksize > maxstacksize)
				maxstacksize = stacksize;
		}*/
		S[l++] = i;
	}
/*	D->stack_size = maxstacksize + 1;*/
	assert(l > 0);
	while(l > 1) {
		last_index = S[--l];
		if (-1 < D->lcp[S[l-1]] && D->lcp[S[l-1]] != D->lcp[last_index]) {
			assert(D->lcp[S[l-1]] < D->lcp[last_index]);
			D->childtab[S[l-1]].down = last_index;
		}
	}
	D->childtab[D->n-1].up = 0;
	my_free(S, D->n * sizeof(int));
}

/* function to get the first l-index of an interval.
   If a singleton interval is given, -1 is returned */
int get_first_l_index(union child *childtab, int left, int right) {
	assert(childtab != NULL);
	if (left == right)
		return -1;
	if (left < childtab[right].up && childtab[right].up<=right)
		return childtab[right].up;
	assert(childtab[left].down>left && childtab[left].down<=right);
	return childtab[left].down;
}

/* function to determine a child interval of the current interval starting with character c
   returns (-1,-1) if there is no such interval
   otherwise, a pair (i,j) is returned representing the child interval.
   the return value will be written to the parameters cleft and cright
   also, the position of the first suffix with a character >= c is returned
   */
int get_interval_starting_with(database *D, int left, int right, unsigned char c, int *cleft, int *cright) {
	int i1, i2;
	unsigned char c2;
	assert(D != NULL && D->childtab != NULL && cleft != NULL && cright != NULL);
	/* if it is a singleton interval, no child interval is left */
	*cleft = *cright = -1;
	if (left == right) {
		return right;
	}
	i1 = get_first_l_index(D->childtab, left,right);
	assert(i1 > left && i1 <= right);
	c2 = D->T[D->SA[left]+D->lcp[i1]];
	if (c2 == c) {
		*cleft = left;
		*cright = i1-1;
		return i1;
	}
	if (c2 > c)
		return left;
	while(i1 <= right) {
		i2 = D->childtab[i1].nextl;
		if (i2 <= i1 || D->lcp[i2] != D->lcp[i1])
			i2 = right+1;
		c2 = D->T[D->SA[i1]+D->lcp[i1]];
		if (c2 == c) {
			*cleft = i1;
			*cright = i2-1;
			return i2;
		}
		if (c2 > c)
			return i1;
		i1 = i2;
	}
	assert(c2 < c && i1 == right+1);
	return i1;
}

void calculate_sl_array(int *qr, database *D) {
	int len, i1, lb1, rb1, lb2, rb2, oldlen;
	memset(D->sl, -1, sizeof(int) * D->n);	
	len = 1;
	D->sl[D->n-1] = 0;
	qr[0] = D->n-1;
	qr[1] = -1;
	while(len > 0) {
		/* push child interval on the stack */
		assert(len <= D->stack_size+2);
		rb1 = qr[len];
		if (rb1 >= 0 && D->sl[rb1] > rb1)
			rb1 = D->sl[rb1];
		lb2 = qr[len-1];
		rb2 = D->sl[lb2];
		if (rb2 < lb2) {
			i1 = lb2;
			lb2 = rb2;
			rb2 = i1;
		}
		assert(lb2 < rb2);
		oldlen = len;
		while (rb1 < rb2) {
			i1 = rb1+1;
			/* first child interval */
			if (i1 == 0) {
				lb1 = lb2;
				rb1 = get_first_l_index(D->childtab, lb2, rb2)-1;
			}
			/* next child intervals */
			else if (D->childtab[i1].nextl > i1 && D->lcp[D->childtab[i1].nextl] == D->lcp[i1]) {
				lb1 = i1;
				rb1 = D->childtab[i1].nextl-1;
			}
			/* last child interval */
			else {
				lb1 = i1;
				rb1 = rb2;
			}
			assert(lb1 <= rb1 && lb2 <= lb1 && rb1 <= rb2 && (lb2 != lb1 || rb1 != rb2));
			/* calculate the suffix link for [lb1, rb1] */
			if (lb1 == rb1)
				continue;
			if (D->sl[lb1] < 0) {
				qr[len] = lb1;
				D->sl[lb1] = rb1;
			}
			else {
				assert(D->sl[rb1] < 0);
				qr[len] = rb1;
				D->sl[rb1] = lb1;
			}
			assert(len <= D->stack_size+1);
			++len;
			qr[len] = -1;
			break;
		}
		if (rb1 >= rb2 && oldlen == len)
			--len;
	}
}

void calculate_suffixlinks(database *D) {
	int *qr, len, i1, l, r, pos, mlcp, rb1, lb1, rb2, lb2, oldlen;
	assert(D != NULL && D->childtab != NULL && D->sl != NULL && D->sr != NULL && D->T != NULL && D->lcp != NULL && D->SA != NULL);
	D->stack_size = D->n;
	qr = my_malloc(sizeof(int) * (D->stack_size + 3));
	calculate_sl_array(qr, D);
	len = 1;
	qr[0] = D->n-1;
	qr[1] = -1;
	i1 = get_first_l_index(D->childtab, 0, D->n-1);
	D->sr[i1] = D->n-1;
	while(len > 0) {
		/* push child interval on the stack */
		assert(len <= D->stack_size+2);
		rb1 = qr[len];
		if (rb1 >= 0 && D->sl[rb1] > rb1)
			rb1 = D->sl[rb1];
		lb2 = qr[len-1];
		rb2 = D->sl[lb2];
		if (rb2 < lb2) {
			i1 = lb2;
			lb2 = rb2;
			rb2 = i1;
		}
		assert(lb2 < rb2);
		oldlen = len;
		while (rb1 < rb2) {
			i1 = rb1+1;
			/* first child interval */
			if (i1 == 0) {
				lb1 = lb2;
				rb1 = get_first_l_index(D->childtab, lb2, rb2)-1;
			}
			/* next child intervals */
			else if (D->childtab[i1].nextl > i1 && D->lcp[D->childtab[i1].nextl] == D->lcp[i1]) {
				lb1 = i1;
				rb1 = D->childtab[i1].nextl-1;
			}
			/* last child interval */
			else {
				lb1 = i1;
				rb1 = rb2;
			}
			assert(lb1 <= rb1 && lb2 <= lb1 && rb1 <= rb2 && (lb2 != lb1 || rb1 != rb2));
			/* calculate the suffix link for [lb1, rb1] */
			if (lb1 == rb1)
				continue;
			if (D->sl[lb1] == rb1) {
				qr[len] = lb1;
			}
			else {
				assert(D->sl[rb1] == lb1);
				qr[len] = rb1;
			}
			pos = get_first_l_index(D->childtab, lb1, rb1);
			assert(pos > lb1 && pos <= rb1);
			mlcp = D->lcp[pos];
			/* get suffix link from parent lcp interval */
			i1 = get_first_l_index(D->childtab, lb2, rb2);
			assert(i1 > lb2 && i1 <= rb2);
			r = D->sr[i1];
			l = D->sl[r];
			if (r < l) {
				i1 = r;
				r = l;
				l = i1;
			}
			assert(l < r);
			while(l < r) {
				unsigned char c;
				i1 = get_first_l_index(D->childtab, l, r);
				assert(i1 > l && i1 <= r);
				if (D->lcp[i1] >= mlcp-1)
					break;
				c = D->T[D->SA[ lb1 ] + D->lcp[i1] + 1];
				get_interval_starting_with(D, l, r, c, &l, &r);
				assert(D->T[D->SA[l]+D->lcp[i1]] == c);
				assert(l >= 0 && l <= r);
			}
			assert(l < r);
			if (D->sl[l] == r)
				D->sr[pos] = l;
			else {
				assert(D->sl[r] == l);
				D->sr[pos] = r;
			}
			assert((D->sr[pos] == l && D->sl[D->sr[pos]] == r) || (D->sr[pos] == r && D->sl[D->sr[pos]] == l));
			/* assert that suffix link is correct */
/*			
			for (i1=0; i1<mlcp-1; ++i1)
				assert(D->T[D->SA[r]+i1] == D->T[D->SA[lb1]+i1+1]);
			if (l > 0) {
				for (i1=0; i1<mlcp-1; ++i1)
					if (D->T[D->SA[l-1]+i1] != D->T[D->SA[l]+i1])
						break;
				assert(i1 < mlcp-1);
			}
			if (r < D->n-1) {
				for (i1=0; i1<mlcp-1; ++i1)
					if (D->T[D->SA[r+1]+i1] != D->T[D->SA[l]+i1])
						break;
				assert(i1 < mlcp-1);
			}
*/
			assert(len <= D->stack_size+1);
			++len;
			qr[len] = -1;
			break;
		}
		if (rb1 >= rb2 && oldlen == len)
			--len;
	}
	my_free(qr, (D->stack_size+3) * sizeof(int));
}

void do_match(database *D, unsigned char *T, int n) {
	int cl, cr, i, l, mlcp, pl, pr, pos, larger;
	unsigned char *DT_ptr;
	FILE *outp;
	assert(D != NULL && D->childtab != NULL && D->sl != NULL && D->sr != NULL && D->T != NULL && D->SA != NULL && D->lcp != NULL);
	assert(T != NULL);
	outp = fopen("merge.tmp", "w");
	if (outp == NULL)
		fatal_error("could not open merge.tmp");
	cl = pl = 0;
	cr = pr = D->n - 1;
	i = l = mlcp = 0;
	while(i < n-1 || l > 0) {
		assert(l >= 0);
		DT_ptr = &D->T[D->SA[cl]+l];
		while(i < n-1 && l < mlcp && *T == *DT_ptr) {
			++i;
			++l;
			++DT_ptr;
			++T;
		}
		assert(l <= mlcp);
		if (l >= mlcp) {
			/* assert it is an internal node */
			assert(l == mlcp);
			while(1) {
				pl = cl;
				pr = cr;
				/* try to match character T[i] */
				larger = get_interval_starting_with(D, cl, cr, *T, &cl, &cr);
				if (cl >= 0) {
					mlcp = (cl == cr? D->n - D->SA[cl] : D->lcp[get_first_l_index(D->childtab, cl, cr)]);
					break;
				}
				fwrite(&larger, 4, 1, outp);
				if (larger > pr)
					cl = l + n + 1;
				else
					cl = l;
				fwrite(&cl, 4, 1, outp);
				if (l == 0) {
					assert(pl == 0 && pr == D->n-1);
					assert(*T != D->T[D->SA[pl]]);
					cl = 0;
					cr = D->n-1;
					++i;
					++T;
					break;
				}
				assert(pl != pr);
				pos = get_first_l_index(D->childtab, pl, pr);
				assert(pos > pl && pos <= pr);
				assert(*T != D->T[D->SA[pl]+l]);
				cl = D->sr[pos];
				if (D->sl[cl] > cl) {
					cr = pr = D->sl[cl];
					pl = cl;
				}
				else {
					assert(D->sl[cl] < cl);
					cr = pr = cl;
					cl = pl = D->sl[cr];
				}
				--mlcp;
				--l;
			}
		}
		else {
			int tl = l;
			if (*T < *DT_ptr) {
				fwrite(&cl, 4, 1, outp);
			}
			else {
				++cr;
				fwrite(&cr, 4, 1, outp);
				--cr;
				tl = l+1+n;
			}
			fwrite(&tl, 4, 1, outp);
			assert(l > 0);
			assert(pl != pr && pl <= cl && pr >= cr);
			pos = get_first_l_index(D->childtab, pl, pr);
			assert(pos > pl && pos <= pr);
			/* use suffix link */
			cl = D->sr[pos];
			if (D->sl[cl] > cl) {
				cr = pr = D->sl[cl];
				pl = cl;
			}
			else {
				assert(D->sl[cl] < cl);
				cr = pr = cl;
				cl = pl = D->sl[cr];
			}
			--l;
			while(cl < cr) {
				pos = get_first_l_index(D->childtab, cl, cr);
				assert(pos > cl && pos <= cr);
				if (D->lcp[pos] >= l) {
					if (D->lcp[pos] == l) {
						pl = cl;
						pr = cr;
					}
					mlcp = D->lcp[pos];
					break;
				}
				pl = cl;
				pr = cr;
				get_interval_starting_with(D, cl, cr, T[D->lcp[pos] - l], &cl, &cr);
				assert(cl >= pl && cr <= pr);
			}
			if (cl == cr) {
				mlcp = D->n - D->SA[cl];
				assert(mlcp >= l);
			}
			else assert(cl < cr);
/*			
			for (pos=0; pos<l; ++pos)
				assert(D->T[D->SA[cl]+pos] == T[pos-l]);
*/			
		}
	}
	fclose(outp);
}
