#include "SAfor2DBnaive.hpp"
using namespace std;

SAfor2DBnaive::SAfor2DBnaive(char* filename1, char* filename2, int format) {
	ConstructSAmanzini constr;
	int overshoot = init_ds_ssort(500, 2000);  // do some initial calculations
	if (overshoot == 0) {
	  cout << "FATAL ERROR IN SUFFIX ARRAY CONSTRUCTION\n";
	  exit(-1);
	}
	SA = construct(&constr, filename1, filename2, format, overshoot);
	if (SA == NULL) {
	  cout << "FATAL ERROR IN SUFFIX ARRAY CONSTRUCTION\n";
	  exit(-1);
	}
}

SAfor2DBnaive::SAfor2DBnaive(char* filename1, char* filename2, 
							 int format, ConstructSA* constr) {
	SA = construct(constr, filename1, filename2, format);
}

int SAfor2DBnaive::index(int i) {
  int li = 0, re = anz, mitte = (re + li) / 2;
  while (bound[mitte] >= i || i > bound[mitte+1]) { // binary search
	if (i <= bound[mitte]) re = mitte;
	else li = mitte;
	mitte = (re + li) / 2;
  }
  return mitte;
}

int SAfor2DBnaive::readln(char* s, int n, FILE* iop) {
  register int c;
  register char* cs;
  register int z = 0; // count characters
  cs=s;
  while (--n > 0 && (c = getc(iop)) != EOF) {
	if ((*cs++ = c) == '\n') break;
	z++;
  }
  if (c == EOF && z==0) z = -1; // signify end of file
  *cs = '\0';
  return z;
}

/*
 * overshoot is only used for the Manzini-method and indicates
 * how much extra space for the text has to be allocated.
 */
int* SAfor2DBnaive::construct(ConstructSA* constr, char* filename1,
							 char* filename2, int format, int overshoot) {
  FILE *fd, *fd2;
  struct stat fdstat, fdstat2;
  long fdpos;
  size_t fdread;
  int stat_res, stat_res2;
  unsigned char* x; // will be used to 'walk' through the text
  char line[FASTA_MAXLINELENGTH];   // for reading the file line by line

  // read input files to memory:
  fd = fopen(filename1, "r");
  stat_res = stat(filename1,&fdstat);
  fd2 = fopen(filename2, "r");
  stat_res2 = stat(filename2,&fdstat2);
  if(fd && !stat_res && fd2 && !stat_res2) {
	switch(format) {
	case BUILTIN:
	  // read first file:
	  length = fdstat.st_size + fdstat2.st_size;
	  middle = fdstat.st_size;
	  t = (unsigned char*) malloc(length + 1 + overshoot);
          change_mem(length + 1 + overshoot);
	  fdpos = 0;
	  fdread = 0;
	  fdread = fread(((char*) t)+fdpos, sizeof(char), fdstat.st_size, fd);
	  fdpos += fdread;
	  fclose(fd);

	  // read second file:
	  fdread = 0;
	  fdread = fread(((char*) t)+fdpos, sizeof(char), fdstat2.st_size, fd2);
	  fdpos += fdread;
	  fclose(fd2);
	  t[fdpos] = '\0';

	  //cout << "file succesfully read!\n";
	  // replace newlines by something very small...
	  x = t;
	  anz = 0; // number of DB-entries
	  anz1 = 0; anz2 = 0;
	  while (*x) {
		if (*x == '\n') {
		  *x = '\1'; anz++;
		  if (x-t < middle) anz1++; else anz2++;
		}
		x++;
	  }
	  break;
	case FASTA:
	  int z;              // length of line that is currently being read
	  length = 0;         // length of string to be constructed
	  anz1 = 0, anz2 = 0; // number of DB-entries

	  // walk through file to calculate the memory needed:
	  while ((z = readln(line, FASTA_MAXLINELENGTH, fd)) >= 0) {
		if (line[0] == '>') {
		  length += 1; // for '\1'-termination
		  anz1 += 1;
		}
		else length += z;
	  }
	  middle = length;

	  // walk through 2nd file:
	  while ((z = readln(line, FASTA_MAXLINELENGTH, fd2)) >= 0) {
		if (line[0] == '>') {
		  length += 1; // for '\1'-termination
		  anz2 += 1;
		}
		else length += z;
	  }
	  anz = anz1 + anz2;
	  t = (unsigned char*) malloc(length + 1 + overshoot);
          change_mem(length + 1 + overshoot);

	  // walk through file once again:
	  rewind(fd);
	  length = 0; // will be re-calculated in this step
	  while (readln(line, FASTA_MAXLINELENGTH, fd) >= 0)
		if (line[0] == '>') break; // jump to first line starting with ">"
	  while ((z = readln(line, FASTA_MAXLINELENGTH, fd)) >= 0) {
		if (line[0] == '>')
		  t[length++] = '\1'; // terminate DB-entry
		else {
		  strncpy(((char*) t) + length, line, z);
		  length += z;
		}
	  }
	  fclose(fd);
	  //cout << "file succesfully read!\n";
	  t[length++] = '\1'; // terminate last DB-entry


	  // walk through 2nd file once again:
	  rewind(fd2);
	  while (readln(line, FASTA_MAXLINELENGTH, fd2) >= 0)
		if (line[0] == '>') break; // jump to first line starting with ">"
	  while ((z = readln(line, FASTA_MAXLINELENGTH, fd2)) >= 0) {
		if (line[0] == '>')
		  t[length++] = '\1'; // terminate DB-entry
		else {
		  strncpy(((char*) t) + length, line, z);
		  length += z;
		}
	  }
	  fclose(fd2);
	  //	  cout << "file succesfully read!\n";
	  t[length++] = '\1'; // terminate last DB-entry
	  t[length] = '\0';   // terminate whole string

	  break;
	default:
	  cout << "Unknown format!\n";
	  return NULL;
	}
  }
  else {
	printf("<--> Error: Could not open %s or %s !\n", filename1, filename2);
	return NULL;
  }


  // now calculate the boundaries of the DB-entries
  bound = new int[anz + 1];
  change_mem(anz*4+4);
  bound[0] = -1;
  x = t;
  int i = 1;
  while (*x) {
	if (*x == '\1') { bound[i++] = x - t; }
	x++;
  }

  // now construct SA (DO include NULL-termination because
  // some construction methods require that (e.g. Larsson's))
  return constr->construct((unsigned char*) t, length+1);
}

SAfor2DBnaive::~SAfor2DBnaive() {
  free(t);
  free(SA);
  delete[] bound;
}
