#include "SAforMDBnaive.hpp"
#include <assert.h>
using namespace std;

SAforMDBnaive::SAforMDBnaive(char* filename) {
	ConstructSAmanzini constr;
	int overshoot = init_ds_ssort(500, 2000);  // do some initial calculations
	if (overshoot == 0) {
	  cout << "FATAL ERROR IN SUFFIX ARRAY CONSTRUCTION\n";
	  exit(-1);
	}
	SA = construct(&constr, filename, overshoot);
	if (SA == NULL) {
	  cout << "FATAL ERROR IN SUFFIX ARRAY CONSTRUCTION\n";
	  exit(-1);
	}
}

int SAforMDBnaive::index(int i) {
  int li = 0, re = anz, mitte = (re + li) / 2;
  while (bound[mitte] >= i || i > bound[mitte+1]) { // binary search
	if (i <= bound[mitte]) re = mitte;
	else li = mitte;
	mitte = (re + li) / 2;
  }
  assert(mitte < anz);
  return mitte;
}

int SAforMDBnaive::readln(char* s, int n, FILE* iop) {
  register int c;
  register char* cs;
  register int z = 0; // count characters
  cs=s;
  while (--n > 0 && (c = getc(iop)) != EOF) {
	if ((*cs++ = c) == '\n') break;
	z++;
  }
  if (c == EOF && z==0) z = -1; // signify end of file
  *cs = '\0';
  return z;
}

/*
 * overshoot is only used for the Manzini-method and indicates
 * how much extra space for the text has to be allocated.
 */
int* SAforMDBnaive::construct(ConstructSA* constr, char* filename, int overshoot) {
  FILE *inif;
  FILE *fd;
  int format;
  char filename1[100];
  struct stat fdstat;
  int stat_res;
  unsigned char* x; // will be used to 'walk' through the text
  char line[FASTA_MAXLINELENGTH];   // for reading the file line by line
  m = 0;
  inif = fopen(filename, "r");
  while(fgets(line, sizeof(line), inif)) ++m;
  // fprintf(stderr, "m = %d\n", m);
  fclose(inif);
  anz2 = new int[m];
  start = new int[m];
  minf = new int[m];
  maxf = new int[m];
  inif = fopen(filename,"r");
  int have = 1000;
  t = (unsigned char *)malloc(1000);
  length = anz = 0;
  for (int i=0; i<m; ++i) {
     fscanf(inif, "%s %d %d %d",filename1, &format, &minf[i], &maxf[i]);
     // printf("trying to read %s in format %d\n", filename1, format);
     // read input files to memory:
     fd = fopen(filename1, "r");
     stat_res = stat(filename1,&fdstat);
     anz2[i] = 0;
     if(fd && !stat_res) {
	int len = 0;
	bool changed = false;
	switch(format) {
	case BUILTIN:
	  start[i] = length;
	  length += fdstat.st_size;
          while(length + 1 + overshoot > have) {
		changed = true;
		have *= 2;
          }
	  if (i+1 == m)
		have = min(have, length + 1 + overshoot);
          if (changed)
	  	t = (unsigned char *)realloc(t, have);
	  // read file:
	  fread(((char*) t)+start[i], sizeof(char), fdstat.st_size, fd);
	  fclose(fd);

	  //cout << "file succesfully read!\n";
	  // replace newlines by something very small...
	  x = t + start[i];
          t[length] = 0;
	  while (*x) {
		if (*x == '\n') {
		  *x = '\1'; anz++;
                  anz2[i]++;
		}
		x++;
	  }
	  break;
	case FASTA:
	  int z;              // length of line that is currently being read

	  // walk through file to calculate the memory needed:
	  while ((z = readln(line, FASTA_MAXLINELENGTH, fd)) >= 0) {
		if (line[0] == '>') {
		  len += 1; // for '\1'-termination
		  anz2[i] += 1;
                  ++anz;
		}
		else len += z;
	  }
	  while(length + len + 1 + overshoot > have) {
		have *= 2;
		changed = true;
	  }
	  if (i+1 == m)
		have = min(have, length + len + 1 + overshoot);
	  if (changed)
		t = (unsigned char *)realloc(t, have);
	  // walk through file once again:
	  rewind(fd);
          start[i] = length;
	  while (readln(line, FASTA_MAXLINELENGTH, fd) >= 0)
		if (line[0] == '>') break; // jump to first line starting with ">"
	  while ((z = readln(line, FASTA_MAXLINELENGTH, fd)) >= 0) {
		if (line[0] == '>')
		  t[length++] = '\1'; // terminate DB-entry
		else {
		  strncpy(((char*) t) + length, line, z);
		  length += z;
		}
	  }
	  fclose(fd);
	  //cout << "file succesfully read!\n";
	  t[length++] = '\1'; // terminate last DB-entry
	  t[length] = '\0';   // terminate whole string

	  break;
	default:
	  cout << "Unknown format!\n";
	  return NULL;
	}
     }
     else {
	printf("<--> Error: Could not open %s!\n", filename1);
	return NULL;
     }
  }
  change_mem(have);
  // now calculate the boundaries of the DB-entries
  bound = new int[anz + 1];
  change_mem(anz*4+4);
  bound[0] = -1;
  x = t;
  int i = 1;
  while (*x) {
	if (*x == '\1') { bound[i++] = x - t; }
	x++;
  }
  // printf("have = %d, length = %d, anz = %d\n", have, length, anz);

  // now construct SA (DO include NULL-termination because
  // some construction methods require that (e.g. Larsson's))
  return constr->construct((unsigned char*) t, length+1);
}

SAforMDBnaive::~SAforMDBnaive() {
  free(t);
  free(SA);
  delete [] start;
  delete [] anz2;
  delete [] minf;
  delete [] maxf;
  delete[] bound;
}
