#ifndef _SAforMDBnaive_hpp_
#define _SAforMDBnaive_hpp_

// constants for readable DB-formats:
#define BUILTIN 1
#define FASTA 2
#define FASTA_MAXLINELENGTH 256

#include "constructSA.hpp"
#include "../globals.h"
#include "manzini.hpp"
#include <stdlib.h>
#include <iostream>
#include <stdio.h>
#include <sys/stat.h>
#include <time.h>
#include <fstream>
using namespace std;

class SAforMDBnaive {
public:
  // construct SA with Manzini's method
  SAforMDBnaive(char*);

  // destructor
  virtual ~SAforMDBnaive();

  // return the index of database entry which occurs in SA at pos. i
  // Currently implemented in time O(log length) instead of O(1)
  // which saves 1 array of size length.
  // DO NOT CALL with Position of 0-byte (i.e., with i=0)
  virtual int index(int);

  // the suffix array (for t)
  int* SA;

  // the text (contains both databases)
  unsigned char* t;

  // number of databases
  int m;

  // number of DB entries in the databases
  int *anz2;

  // total number of DB entries in all DBs
  int anz;

  // total length of DB (i.e., the SA has length+1 entries (0-byte))
  int length;

  // start positions where databases start
  int *start;

  // minf
  int *minf;

  // maxf
  int *maxf;

private:
  // method for constructing the SA (reads complete DB)
  int* construct(ConstructSA*, char*, int);

  // like fgets, but return number of chars
  int readln(char*, int, FILE*);

  // saves borders of DB-entries
  int* bound;
};

#endif
