#include "frequent_linear2.hpp"
#include <assert.h>

Node::Node(int l, int *s, int m) {
	this->l = l;
	this->s = new int[m];
	for (int i=0; i<m; ++i) {
                if (s != NULL) this->s[i] = s[i];
                else this->s[i] = 0;
	}
}

void FrequentStringsLinear::print(int* SA, unsigned char* t, int p, int l) {
  for (int i = SA[p]; (i < SA[p] + l) && ((char) t[i] > 1); i++)
	cout << (char) t[i];
  cout << endl;
}

#include <time.h>

// This method assumes the LCP-array is LCP[0..length] and
// LCP[length] = LCP[0] = -1. The SA is SA[0..length-1].
void FrequentStringsLinear::frequent(char* filename) {
  FILE *progress = fopen("progress.txt","w");
  int start = clock();
  /**
   * FIRST PHASE: PREPROCESSING
   **/
  SAforMDBnaive* SA4DB = new SAforMDBnaive(filename);
  int length = SA4DB->length + 1; // length of LCP (not counting last -1)
  LCPmanzini manzini(SA4DB->SA, SA4DB->t, length);
  int* LCP = manzini.constructDB();
  RMQ_n_1_alstrup* rmq = new RMQ_n_1_alstrup(LCP, length+1);
  int cur = clock();
  fprintf(progress, "preprocess done in %.2lf seconds\n", (double)(cur-start)/CLOCKS_PER_SEC);
  fflush(progress);
  start = cur;

  /**
   * SECOND PHASE: LABELLING
   **/
  int **count = new int*[SA4DB->m];
  for (int i=0; i<SA4DB->m; ++i) {
	count[i] = new int[length];
	memset(count[i], 0, 4 * length);
  }
  change_mem(length * 4LL * SA4DB->m);
  int* last = new int[SA4DB->anz];
  change_mem(4LL * SA4DB->anz);
  for (int i = 0; i < SA4DB->anz; i++) last[i] = 0;

  int entry; // # of entry in DB
  int pos;   // position of RMQ
  for (int i = 1; i < length; i++) {
	entry = SA4DB->index(SA4DB->SA[i]);
	if (last[entry] != 0) {
	  pos = rmq->query(last[entry]+1,i);
      int id, rem = entry;
      /* get database id */
	  for (id = 0; id<SA4DB->m; ++id) {
              if (SA4DB->anz2[id] > rem) break;
              rem -= SA4DB->anz2[id];
      }
      *(count[id] + pos) =  *(count[id] + pos) + 1;
	}
	last[entry] = i;
  }
  fprintf(progress, "deleting rmq\n");
  fflush(progress);
  delete rmq;
  change_mem(-4LL * SA4DB->anz);
  delete[] last;
  for (int i=0; i<SA4DB->m; ++i) {
	for (int j=1; j<length; ++j)
		*(count[i]+j) = *(count[i]+j) + *(count[i]+j-1);
  }
  cur = clock();
  fprintf(progress, "correction terms done in %.2lf seconds\n", (double)(cur-start)/CLOCKS_PER_SEC);
  fflush(progress);
  start = cur;

  /**
   * THIRD PHASE: EXTRACTION
   **/
  stack<Node*> S;             // stack for DF-traversal (keeps nodes)
  Node *node, *node2;         // variables needed for the DF-traversal
  int left;                   // first index of string in SA
  int *olds = new int[SA4DB->m]; // s-values of last completely visited subtree
  int h_parent;               // stores height of parent node (for output)
  int freq;
  int h, nexth;
  S.push(new Node(0, NULL, SA4DB->m));    // virtual root (acts as stopper)
  change_mem((SA4DB->m + 3)*4);
  LCP[0] = LCP[length] = 0;
  for (int k = 1; k <= length; k++) { // start at 1 to skip the "$"!
	node = S.top(); // go to next leaf
	left = k-1;
        memset(olds, 0, sizeof(int) * SA4DB->m);
        if (k == 1)
		h = 0;
	else
		h = SA4DB->length - SA4DB->SA[k-1];
	while (h > LCP[k]) {
	  left = node->l;   // update left border
          nexth = LCP[left];
	  S.pop();              // the popped node is now in 'node'

      // propagate counters upwards:
	  node2 = S.top();
	  if (nexth >= LCP[k]) { // otherwise top-node not LCA
                for (int i=0; i<SA4DB->m; ++i)
			node2->s[i] += node->s[i];
		h_parent = nexth; // for outputting all patterns
	  }
	  else h_parent = LCP[k];
	  // change the following if-statements according to the strings of interest
          bool valid = SA4DB->t[SA4DB->SA[left]+nexth] > '\1' && SA4DB->t[SA4DB->SA[left]+LCP[k]] > '\1';
	  for (int i=0; valid && i<SA4DB->m; ++i) {
		freq = node->s[i] - (*(count[i]+k-1)-*(count[i]+left));
		if (freq < SA4DB->minf[i] || freq > SA4DB->maxf[i])
			valid = false;
	  }

   	  if (valid) {
		// calculate true length of string s to be printed
		// (time for loop is O(|s|))
		for (int h2 = 0; h2 < h; h2++) {
		  if (SA4DB->t[SA4DB->SA[left]+h2] <= '\1') {
			h = h2; break;
		  }
		}
		for (int h2 = h_parent + 1; h2 <= h; h2++) {
		  print(SA4DB->SA, SA4DB->t, left, h2);
		}
	  }
	  h = nexth;
          for (int i=0; i<SA4DB->m; ++i)
		olds[i] = node->s[i];
	  delete node;      // free space
          change_mem(-(SA4DB->m+3)*4);
	  node = S.top();   // next node on stack
	}
	if (h < LCP[k]) { // push lca (every inner node IS lca of consec. leaves!)
	  S.push(new Node(left, olds, SA4DB->m));
          change_mem((SA4DB->m + 3) * 4);
	}

    // push leaf:
	if (k < length) {
	  entry = SA4DB->index(SA4DB->SA[k]); // check if current string is from DB1 or DB2
	  S.push(new Node(k, NULL, SA4DB->m));
          for (int id=0; id<SA4DB->m; ++id) {
		if (SA4DB->anz2[id] > entry) {
                        S.top()->s[id] = 1;
			break;
		}
		entry -= SA4DB->anz2[id];
	  }
          change_mem((SA4DB->m + 3) * 4);
	}
  }
  cur = clock();
  fprintf(progress, "extraction phase done in %.2lf seconds\n", (double)(cur-start)/CLOCKS_PER_SEC);
  fprintf(progress, "maximum  memory consumption: %lld\n", maxmem);
  fflush(progress);
  delete[] LCP;
  for (int i=0; i<SA4DB->m; ++i)
	delete [] count[i];
  delete[] count;
  delete SA4DB;
  assert(!S.empty());
  delete S.top();
}
