#include "globals.h"
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <malloc.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

const unsigned char SEPARATOR = 1;
long long maxmem = 0;
long long curmem = 0;
int overshoot;
int start_time = -1;

void change_mem(long long diff) {
	curmem += diff;
	if (curmem > maxmem)
		maxmem = curmem;
}

void fatal_error(const char *message) {
	puts(message);
	exit(1);
}

int MAX(int a, int b) {
	return a > b? a : b;
}

int MIN(int a, int b) {
	return a < b? a : b;
}

/*
void generate_random_filename(char *s) {
	int i;
	for (i=0; i<5; ++i)
		s[i] = rand()%26 + 'a';
	s[5] = '.';
	s[6] = 't';
	s[7] = 'm';
	s[8] = 'p';
	s[9] = 0;
}
*/

int my_open(char *fname, int flags, int mode) {
	int fd;
	fd = open(fname, flags, mode);
	if (fd < 0) {
		perror("open file");
		exit(1);
	}
	return fd;
}

int open_truncate(char *prefix, char *dbname, char *fname, int size) {
	int fd;
	sprintf(fname, "%s%s", prefix, dbname);
/*	generate_random_filename(fname); */
	fd = my_open(fname, O_RDWR | O_CREAT, 0600);
 	if (ftruncate(fd, size) < 0) {
		perror("ftruncate file");
		exit(1);
	}
	return fd;
}

void my_close(int fd) {
	if (close(fd) < 0) {
		perror("close file");
		exit(1);
	}
}

void close_unlink(int fd, char *fname) {
	my_close(fd);
	if (unlink(fname) < 0) {
		perror("unlink file");
		exit(1);
	}
}

int *my_malloc(int size) {
	int *r;
	change_mem(size);
	r = (int *)malloc(size);
	if (r == NULL)
		fatal_error("not enough memory");
	return r;
}

int *my_calloc(int n, int size) {
	int *r;
	change_mem((long long)size * n);
	r = (int *)calloc(n, size);
	if (r == NULL)
		fatal_error("not enough memory");
	return r;
}

void my_free(void *ptr, int size) {
	free(ptr);
	change_mem(-size);
}

FILE *progress;

int *my_mmap(int fd, int size, int mode) {
	int *res = (int *)mmap(NULL, size, PROT_READ | PROT_WRITE, mode, fd, 0);
	if (res == MAP_FAILED) {
		perror("mmap");
		exit(1);
	}
	curmem += size;
	if (curmem > maxmem) {
		maxmem = curmem;
	}
	return res;
}

void my_msync(void *ptr, int size) {
	if (msync(ptr, size, MS_SYNC) < 0) {
		perror("msync");
		exit(1);
	}
}

void my_munmap(void *ptr, int size) {
	curmem -= size;
	if (munmap(ptr, size) < 0) {
		perror("munmap");
		exit(1);
	}
}

void print_progress(const char *s) {
	int next_time = clock();
	if (start_time >= 0) {
		fprintf(progress, "%s done in %.2lf seconds\n", s, (next_time-start_time)/(double)CLOCKS_PER_SEC);
		fflush(progress);
	}
	else {
		progress = fopen("progress.txt","w");
		if (progress == NULL)
			fatal_error("could not open progress.txt");
	}
	start_time = next_time;
}
