#include "preprocess.h"
#include "datastructures.h"
#include "extract.h"
#include "correction_terms.h"
#include "globals.h"
#include "merge.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

extern int curmem;
extern int overshoot;
extern FILE *progress;

void rewrite(database *D, int n, int *c) {
        FILE *inp, *outp;
        int i, v, p, *ISA, *a;
        assert(D != NULL && D->lcp != NULL && D->SA != NULL && c != NULL);
        my_msync(D->lcp, D->n * sizeof(int));
        my_munmap(D->lcp, D->n * sizeof(int));
		ISA = my_malloc(D->n * sizeof(int));
        D->lcp = NULL;
        for (i=0; i<D->n; ++i)
                ISA[D->SA[i]] = i;
        inp = fopen("merge.tmp", "r");
        if (inp == NULL)
                fatal_error("could not open merge.tmp");
		a = my_malloc(D->n * sizeof(int));
		memset(c, 0, sizeof(int) * (n+1));
        for (i=0; i<D->n-1; ++i) {
				fread(&p, 4, 1, inp);
				assert(p > 0 && p <= n);
				++c[n-p];
                fread(&v, 4, 1, inp);
                a[ISA[i]] = v;
        }
		++c[n-1];
        a[ISA[D->n-1]] = 0;
        fclose(inp);
		my_free(ISA, D->n * sizeof(int));
        outp = fopen("merge.tmp", "w");
        if (outp == NULL)
                fatal_error("could not open merge.tmp");
        for (i=D->n-1; i>=0; --i)
                fwrite(&a[i], 4, 1, outp);
		my_free(a, D->n * sizeof(int));
        fclose(outp);
        D->lcp = my_mmap(D->lcp_fd, D->n * sizeof(int), MAP_PRIVATE);
}

#ifdef MERGE_ONLY
void parse_inifile(char *filename) {
	char buf[100], name[100];
	database res, cur;
	FILE *inifile = fopen(filename, "r");
	if (inifile == NULL) {
		perror("parse_inifile");
		exit(1);
	}
	res.n = 0;
	res.c = cur.c = NULL;
	while(fgets(buf, sizeof(buf), inifile)) {
		if (sscanf(buf, "%s %d %d", name, &cur.minf, &cur.maxf) != 3)
			fatal_error("expected a line <filename> <minf> <maxf>");
		strcpy(cur.db_name, name);
		read_database(name, &cur);
		sprintf(cur.a_filename, "a_%s", name);
		sprintf(cur.b_filename, "b_%s", name);
		sprintf(cur.sa_filename, "sa_%s", name);
		sprintf(cur.lcp_filename, "lcp_%s", name);
		cur.a_fd = my_open(cur.a_filename, O_RDWR, 0600);
		cur.b_fd = my_open(cur.b_filename, O_RDWR, 0600);
		cur.sa_fd = my_open(cur.sa_filename, O_RDWR, 0600);
		cur.lcp_fd = my_open(cur.lcp_filename, O_RDWR, 0600);
		if (cur.minf < 0)
			fatal_error("minf should be >= 0");
		if (cur.minf > cur.maxf)
			fatal_error("maxf should be >= minf");
		if (res.n != 0) {
			if (res.a != NULL) {
				my_msync(res.a, res.n * sizeof(int));
				my_munmap(res.a, res.n * sizeof(int));
				res.a = NULL;
			}
			if (res.b != NULL) {
				my_msync(res.b, res.n * sizeof(int));
				my_munmap(res.b, res.n * sizeof(int));
				res.b = NULL;
			}
			if (res.SA == NULL)
				res.SA = my_mmap(res.sa_fd, res.n * sizeof(int), MAP_PRIVATE);
			if (res.sl == NULL)
				res.sl = my_mmap(res.sl_fd, res.n * sizeof(int), MAP_PRIVATE);
			if (res.sr == NULL)
				res.sr = my_mmap(res.sr_fd, res.n * sizeof(int), MAP_PRIVATE);
			if (res.childtab == NULL)
				res.childtab = (union child *)my_mmap(res.ch_fd, res.n * sizeof(int), MAP_PRIVATE);
			if (res.T == NULL)
				res.T = (unsigned char *)my_mmap(res.t_fd, res.n + overshoot, MAP_PRIVATE);
			assert(res.c == NULL && res.lcp != NULL);
			do_match(&res, cur.T, cur.n);
			print_progress("do_match");

			my_msync(res.sl, res.n * sizeof(int));
			my_munmap(res.sl, res.n * sizeof(int));
			my_msync(res.sr, res.n * sizeof(int));
			my_munmap(res.sr, res.n * sizeof(int));
			my_msync(res.childtab, res.n * sizeof(int));
			my_munmap(res.childtab, res.n * sizeof(union child));
			my_msync(res.T, res.n + overshoot);
			my_munmap(res.T, res.n + overshoot);
			res.sl = res.sr = NULL;
			res.childtab = NULL;
			res.T = NULL;
			assert(res.SA != NULL);
			my_msync(res.SA, sizeof(int) * res.n);
			my_munmap(res.SA, sizeof(int) * res.n);
			res.SA = NULL;
			assert(res.lcp != NULL);
			my_msync(res.lcp, sizeof(int) * res.n);
			my_munmap(res.lcp, sizeof(int) * res.n);
			res.lcp = NULL;
		}
		assert(cur.T != NULL);
		my_msync(cur.T, cur.n + overshoot);
		my_munmap(cur.T, cur.n + overshoot);
		cur.T = NULL;
		cur.SA = my_mmap(cur.sa_fd, cur.n * sizeof(int), MAP_SHARED);
		cur.lcp = my_mmap(cur.lcp_fd, cur.n * sizeof(int), MAP_SHARED);
		cur.a = NULL;
		cur.b = NULL;
		if (res.n != 0) {
			res.c = my_mmap(res.c_fd, (res.n+1) * sizeof(int), MAP_SHARED);
			rewrite(&cur, res.n, res.c);
			print_progress("rewrite");
			my_msync(res.c, (res.n+1)*sizeof(int));
			my_munmap(res.c, (res.n+1)*sizeof(int));
			res.c = NULL;
			assert(cur.SA != NULL);
			my_msync(cur.SA, cur.n * sizeof(int));
			my_munmap(cur.SA, cur.n * sizeof(int));
			cur.SA = NULL;
			assert(cur.lcp != NULL);
			my_msync(cur.lcp, cur.n * sizeof(int));
			my_munmap(cur.lcp, cur.n * sizeof(int));
			cur.lcp = NULL;
			assert(cur.T == NULL);
			assert(cur.a == NULL && cur.b == NULL);
			cur.a = my_mmap(cur.a_fd, cur.n * sizeof(int), MAP_SHARED);
			cur.b = my_mmap(cur.b_fd, cur.n * sizeof(int), MAP_SHARED);
			res.a = my_mmap(res.a_fd, res.n * sizeof(int), MAP_SHARED);
			res.b = my_mmap(res.b_fd, res.n * sizeof(int), MAP_SHARED);
			res.lcp = my_mmap(res.lcp_fd, res.n * sizeof(int), MAP_SHARED);
			merge(&res, &cur);
			print_progress("merge");
			my_msync(res.a, res.n * sizeof(int));
			my_munmap(res.a, res.n * sizeof(int));
			my_msync(res.b, res.n * sizeof(int));
			my_munmap(res.b, res.n * sizeof(int));
			res.a = res.b = NULL;
			res.c = NULL;
			close_database(&cur);
		}
		else {
			res = cur;
			res.T = (unsigned char *)my_mmap(res.t_fd, res.n + overshoot, MAP_PRIVATE);
			assert(res.a == NULL && res.b == NULL);
			assert(res.T != NULL && res.lcp != NULL && res.SA != NULL);
			res.c_fd = open_truncate("c_", res.db_name, res.c_filename, (res.n+1) * sizeof(int));
			res.sl_fd = open_truncate("sl_", res.db_name, res.sl_filename, res.n * sizeof(int));
			res.sr_fd = open_truncate("sr_", res.db_name, res.sr_filename, res.n * sizeof(int));
			res.ch_fd = open_truncate("ch_", res.db_name, res.ch_filename, res.n * sizeof(int));
			res.childtab = (union child *)my_mmap(res.ch_fd, res.n * sizeof(int), MAP_SHARED);
			create_childtab(&res);
			print_progress("calculate childtab");
			res.sl = my_mmap(res.sl_fd, res.n * sizeof(int), MAP_SHARED);
			res.sr = my_mmap(res.sr_fd, res.n * sizeof(int), MAP_SHARED);
			calculate_suffixlinks(&res);
			print_progress("calculate suffixlinks");
			my_msync(res.childtab, res.n * sizeof(int));
			my_munmap(res.childtab, res.n * sizeof(int));
			my_msync(res.sl, res.n * sizeof(int));
			my_munmap(res.sl, res.n * sizeof(int));
			my_msync(res.sr, res.n * sizeof(int));
			my_munmap(res.sr, res.n * sizeof(int));
			res.sl = res.sr = NULL;
			res.childtab = NULL;
		}
	}
	if (res.sl != NULL) {
		my_msync(res.sl, res.n * sizeof(int));
		my_munmap(res.sl, res.n * sizeof(int));
	}
	if (res.sr != NULL) {
		my_msync(res.sr, res.n * sizeof(int));
		my_munmap(res.sr, res.n * sizeof(int));
	}
	if (res.childtab != NULL) {
		my_msync(res.childtab, res.n * sizeof(int));
		my_munmap(res.childtab, res.n * sizeof(int));
	}
	assert(res.c == NULL);
	close_unlink(res.c_fd, res.c_filename);
	close_unlink(res.ch_fd, res.ch_filename);
	close_unlink(res.sl_fd, res.sl_filename);
	close_unlink(res.sr_fd, res.sr_filename);
	if (res.T == NULL)
		res.T = (unsigned char *)my_mmap(res.t_fd, res.n + overshoot, MAP_PRIVATE);
	if (res.b == NULL)
		res.b = my_mmap(res.b_fd, res.n * sizeof(int), MAP_SHARED);
	if (res.SA == NULL)
		res.SA = my_mmap(res.sa_fd, res.n * sizeof(int), MAP_PRIVATE);
	if (res.a == NULL)
		res.a = my_mmap(res.a_fd, res.n * sizeof(int), MAP_PRIVATE);
	print_results(&res);
	close_database(&res);
	print_progress("print results");
	fclose(inifile);
	assert(curmem == 0);
}

#else

void parse_inifile(char *filename) {
	char buf[100], name[100];
	database res, cur;
	int *C;
	FILE *inifile = fopen(filename, "r");
	if (inifile == NULL) {
		perror("parse_inifile");
		exit(1);
	}
	res.n = 0;
	res.c = cur.c = NULL;
	while(fgets(buf, sizeof(buf), inifile)) {
		if (sscanf(buf, "%s %d %d", name, &cur.minf, &cur.maxf) != 3)
			fatal_error("expected a line <filename> <minf> <maxf>");
		strcpy(cur.db_name, name);
		if (cur.minf < 0)
			fatal_error("minf should be >= 0");
		if (cur.minf > cur.maxf)
			fatal_error("maxf should be >= minf");
		read_database(name, &cur);
		if (res.n != 0) {
			if (res.a != NULL) {
				my_msync(res.a, res.n * sizeof(int));
				my_munmap(res.a, res.n * sizeof(int));
				res.a = NULL;
			}
			if (res.b != NULL) {
				my_msync(res.b, res.n * sizeof(int));
				my_munmap(res.b, res.n * sizeof(int));
				res.b = NULL;
			}
			if (res.SA == NULL)
				res.SA = my_mmap(res.sa_fd, res.n * sizeof(int), MAP_PRIVATE);
			if (res.sl == NULL)
				res.sl = my_mmap(res.sl_fd, res.n * sizeof(int), MAP_PRIVATE);
			if (res.sr == NULL)
				res.sr = my_mmap(res.sr_fd, res.n * sizeof(int), MAP_PRIVATE);
			if (res.childtab == NULL)
				res.childtab = (union child *)my_mmap(res.ch_fd, res.n * sizeof(int), MAP_PRIVATE);
			if (res.T == NULL)
				res.T = (unsigned char *)my_mmap(res.t_fd, res.n + overshoot, MAP_PRIVATE);
			assert(res.c == NULL && res.lcp != NULL);
			do_match(&res, cur.T, cur.n);
			print_progress("do_match");

			my_msync(res.sl, res.n * sizeof(int));
			my_munmap(res.sl, res.n * sizeof(int));
			my_msync(res.sr, res.n * sizeof(int));
			my_munmap(res.sr, res.n * sizeof(int));
			my_msync(res.childtab, res.n * sizeof(int));
			my_munmap(res.childtab, res.n * sizeof(union child));
			my_msync(res.T, res.n + overshoot);
			my_munmap(res.T, res.n + overshoot);
			res.sl = res.sr = NULL;
			res.childtab = NULL;
			res.T = NULL;
			assert(res.SA != NULL);
			my_msync(res.SA, sizeof(int) * res.n);
			my_munmap(res.SA, sizeof(int) * res.n);
			res.SA = NULL;
			assert(res.lcp != NULL);
			my_msync(res.lcp, sizeof(int) * res.n);
			my_munmap(res.lcp, sizeof(int) * res.n);
			res.lcp = NULL;
		}
		preprocess(&cur);
		print_progress("preprocess");
		if (res.n != 0) {
			res.c = my_mmap(res.c_fd, (res.n+1) * sizeof(int), MAP_SHARED);
			my_munmap(cur.a, cur.n * sizeof(int));
			rewrite(&cur, res.n, res.c);
			print_progress("rewrite");
			cur.a = my_mmap(cur.a_fd, cur.n * sizeof(int), MAP_SHARED);
			my_msync(res.c, (res.n+1)*sizeof(int));
			my_munmap(res.c, (res.n+1)*sizeof(int));
			res.c = NULL;
		}
		C = calculate_c(&cur);
		print_progress("calculate correction terms");
		extract_relevant_substrings(&cur, C);
		print_progress("extract relevant substrings");
		my_free(C, cur.n * sizeof(int));
		if (res.n != 0) {
			assert(cur.SA != NULL);
			my_msync(cur.SA, cur.n * sizeof(int));
			my_munmap(cur.SA, cur.n * sizeof(int));
			cur.SA = NULL;
			my_munmap(cur.lcp, cur.n * sizeof(int));
			cur.lcp = NULL;
			assert(cur.T == NULL);
			res.a = my_mmap(res.a_fd, res.n * sizeof(int), MAP_SHARED);
			res.b = my_mmap(res.b_fd, res.n * sizeof(int), MAP_SHARED);
			res.lcp = my_mmap(res.lcp_fd, res.n * sizeof(int), MAP_SHARED);
			merge(&res, &cur);
			print_progress("merge");
			my_msync(res.a, res.n * sizeof(int));
			my_munmap(res.a, res.n * sizeof(int));
			my_msync(res.b, res.n * sizeof(int));
			my_munmap(res.b, res.n * sizeof(int));
			res.a = res.b = NULL;
			res.c = NULL;
			delete_database(&cur);
		}
		else {
			res = cur;
			res.T = (unsigned char *)my_mmap(res.t_fd, res.n + overshoot, MAP_PRIVATE);
			assert(res.a != NULL);
			my_msync(res.a, sizeof(int) * res.n);
			my_munmap(res.a, sizeof(int) * res.n);
			res.a = NULL;
			assert(res.b != NULL);
			my_msync(res.b, sizeof(int) * res.n);
			my_munmap(res.b, sizeof(int) * res.n);
			res.b = NULL;
			assert(res.T != NULL && res.lcp != NULL && res.SA != NULL);
			res.c_fd = open_truncate("c_", res.db_name, res.c_filename, (res.n+1) * sizeof(int));
			res.sl_fd = open_truncate("sl_", res.db_name, res.sl_filename, res.n * sizeof(int));
			res.sr_fd = open_truncate("sr_", res.db_name, res.sr_filename, res.n * sizeof(int));
			res.ch_fd = open_truncate("ch_", res.db_name, res.ch_filename, res.n * sizeof(int));
			res.childtab = (union child *)my_mmap(res.ch_fd, res.n * sizeof(int), MAP_SHARED);
			create_childtab(&res);
			print_progress("calculate childtab");
			res.sl = my_mmap(res.sl_fd, res.n * sizeof(int), MAP_SHARED);
			res.sr = my_mmap(res.sr_fd, res.n * sizeof(int), MAP_SHARED);
			calculate_suffixlinks(&res);
			print_progress("calculate suffixlinks");
			my_msync(res.childtab, res.n * sizeof(int));
			my_munmap(res.childtab, res.n * sizeof(int));
			my_msync(res.sl, res.n * sizeof(int));
			my_munmap(res.sl, res.n * sizeof(int));
			my_msync(res.sr, res.n * sizeof(int));
			my_munmap(res.sr, res.n * sizeof(int));
			res.sl = res.sr = NULL;
			res.childtab = NULL;
		}
	}
	if (res.sl != NULL) {
		my_msync(res.sl, res.n * sizeof(int));
		my_munmap(res.sl, res.n * sizeof(int));
	}
	if (res.sr != NULL) {
		my_msync(res.sr, res.n * sizeof(int));
		my_munmap(res.sr, res.n * sizeof(int));
	}
	if (res.childtab != NULL) {
		my_msync(res.childtab, res.n * sizeof(int));
		my_munmap(res.childtab, res.n * sizeof(int));
	}
	assert(res.c == NULL);
	close_unlink(res.c_fd, res.c_filename);
	close_unlink(res.ch_fd, res.ch_filename);
	close_unlink(res.sl_fd, res.sl_filename);
	close_unlink(res.sr_fd, res.sr_filename);
	if (res.T == NULL)
		res.T = (unsigned char *)my_mmap(res.t_fd, res.n + overshoot, MAP_PRIVATE);
	if (res.b == NULL)
		res.b = my_mmap(res.b_fd, res.n * sizeof(int), MAP_SHARED);
	if (res.SA == NULL)
		res.SA = my_mmap(res.sa_fd, res.n * sizeof(int), MAP_PRIVATE);
	if (res.a == NULL)
		res.a = my_mmap(res.a_fd, res.n * sizeof(int), MAP_PRIVATE);
	print_results(&res);
	print_progress("print results");
	delete_database(&res);
	fclose(inifile);
	assert(curmem == 0);
}
#endif

int main(int argc, char **argv) {

#ifdef EXTRACT_ONLY
	database cur;
	int *C;
	print_progress("");
	if (argc != 4) {
		printf("Usage: %s <databasefile> <minf> <maxf>\n", argv[0]);
		exit(1);
	}
	cur.c = NULL;
	if (sscanf(argv[2], "%d", &cur.minf) != 1)
		fatal_error("expected integer <minf> as second parameter");
	if (sscanf(argv[3], "%d", &cur.maxf) != 1)
		fatal_error("expected integer <maxf> as second parameter");
	if (cur.minf < 0)
		fatal_error("minf should be >= 0");
	if (cur.minf > cur.maxf)
		fatal_error("maxf should be >= minf");
	strcpy(cur.db_name, argv[1]);
	read_database(cur.db_name, &cur);
	preprocess(&cur);
	print_progress("preprocess");
	C = calculate_c(&cur);
	print_progress("calculate correction terms");
	extract_relevant_substrings(&cur, C);
	print_progress("extract relevant substrings");
	my_free(C, cur.n * sizeof(int));
	/*
	if (cur.SA == NULL)
		cur.SA = my_mmap(cur.sa_fd, cur.n * sizeof(int), MAP_PRIVATE);
	if (cur.T == NULL)
		cur.T = (unsigned char *)my_mmap(cur.t_fd, cur.n + overshoot, MAP_PRIVATE);
	print_results(&cur);
	*/
	close_database(&cur);
	fprintf(progress, "maximum memory usage: %lld\n", maxmem);
	fclose(progress);
	assert(curmem == 0);
	return 0;
#endif

	if (argc != 2) {
		printf("Usage: %s <initfile>\n", argv[0]);
		printf("initfile should contain one line for each database\n");
		printf("lines are formatted as follows: <filename> <minf> <maxf>\n");
		exit(1);
	}

	print_progress("");
	parse_inifile(argv[1]);
	fprintf(progress, "maximum memory usage: %lld\n", maxmem);
	fclose(progress);
	return 0;
}
