#include "preprocess.h"
#include "globals.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>

void construct_lcparray(database *D) {
	int i, j, k, l = 0;
	D->lcp_fd = open_truncate("lcp_", D->db_name, D->lcp_filename, D->n * sizeof(int));
	D->lcp = my_mmap(D->lcp_fd, sizeof(int) * D->n, MAP_SHARED);
	/* use D->a for inverse suffix array */
	for (i=0; i<D->n; ++i)
		D->a[D->SA[i]] = i;
	D->lcp[0] = 0;
	for (i=0; i<D->n; ++i) {
		k = D->a[i];
		if (k > 0) {
			j = D->SA[k-1];
			while(i+l < D->n && j+l < D->n && D->T[i+l] == D->T[j+l])
				++l;
			D->lcp[k] = l;
		}
		if (l > 0)
			--l;
	}
}

extern int overshoot;

/* open file and read the data */
void read_database(char *filename, database *D) {
	int i;
	FILE *input = fopen(filename, "r");
	if (input == NULL) {
		perror("preprocess");
		exit(1);
	}
	assert(D != NULL);
	overshoot = init_ds_ssort(500, 2000);
	if (overshoot == 0)
		fatal_error("ds initialization failed! (preprocess)");
	if (fseek(input, 0, SEEK_END) != 0) {
		perror("preprocess");
		exit(1);
	}
	D->n = ftell(input);
	if (D->n < 0) {
		perror("D->n");
		exit(1);
	}
	if (D->n == 0)
		fatal_error("Input file empty");
	D->t_fd = open_truncate("t_", D->db_name, D->t_filename, D->n + overshoot + 1);
	D->T = (unsigned char *)my_mmap(D->t_fd, D->n + overshoot + 1, MAP_SHARED);
	rewind(input);
	if ((int)fread(D->T, 1, D->n, input) != D->n) {
		perror("reading input file");
		exit(1);
	}
	for (i=0; i<D->n; ++i) {
		if (D->T[i] == '\n')
			D->T[i] = SEPARATOR;
		else
			assert(D->T[i] > SEPARATOR);
	}
	D->T[D->n] = 0;
	++D->n;
	fclose(input);
}

/* preprocess the data */
void preprocess(database *D) {
	/* int i; */
	/* construct the suffix array */
	D->sa_fd = open_truncate("sa_", D->db_name, D->sa_filename, D->n * sizeof(int));
	D->SA = my_mmap(D->sa_fd, D->n * sizeof(int), MAP_SHARED);
	ds_ssort(D->T, D->SA, D->n);
	/* allocate memory using mmap for arrays a and b */
	D->a_fd = open_truncate("a_", D->db_name, D->a_filename, D->n * sizeof(int));
	D->a = my_mmap(D->a_fd, D->n * sizeof(int), MAP_SHARED);
	construct_lcparray(D);
}
